import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import java.util.*;
import java.io.*;
import javax.microedition.rms.*;

public class BookReader extends MIDlet
{
    static final Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 1);
    static final Command QUIT_CMD = new Command(Locale.QUIT, Command.CANCEL, 2);
    static final Command BOOKMARKS_CMD = new Command(Locale.BOOKMARKS, Command.SCREEN, 1);
    static final Command GOTO_CMD = new Command(Locale.GOTO, Command.SCREEN, 1);
    static final Command ADD_CMD = new Command(Locale.ADD, Command.SCREEN, 2);
    static final Command RENAME_CMD = new Command(Locale.RENAME, Command.SCREEN, 1);
    static final Command REMOVE_CMD = new Command(Locale.REMOVE, Command.SCREEN, 3);
    static final Command SEARCH_CMD = new Command(Locale.SEARCH, Command.SCREEN, 2);
    static final Command SETTINGS_CMD = new Command(Locale.SETTINGS, Command.SCREEN, 3);
    static final Command SAVE_CMD = new Command(Locale.SAVE, Command.SCREEN, 1);
    static final Command CLOCK_CMD = new Command(Locale.CLOCK, Command.SCREEN, 4);
    static final Command HELP_CMD = new Command(Locale.HELP, Command.SCREEN, 5);

    static final String BOOK = "/book.txt";
    static final String STORE_NAME = "BOOK_READER";

    static final int DEFAULT_OVERLAP = 1;

    static final int BUF_SIZE = 8*1024;

    private int insertBookmark(Bookmark bm) { 
        int pos = bm.pos;
        int l = 0, n = bookmarks.size(), r = n;
        while (l < r) { 
            int m = (l + r) >> 1;
            if (pos > ((Bookmark)bookmarks.elementAt(m)).pos) { 
                l = m + 1;
            } else { 
                r = m;
            }
        }
        bookmarks.insertElementAt(bm, r);
        return r;
    }

    public int search(int fromPos, String pattern) { 
        char[] p = pattern.toLowerCase().toCharArray();
        int n  = p.length;
        char[] q = new char[n];
        char ch, prev = '\0';
        int i = 0;
      search:
        while ((ch = getChar(fromPos++)) != '\0') { 
            if (ch ==  ' ' && prev == ' ') { 
                continue;
            }
            prev = ch;
            if (i == n) { 
                i = 0;
            }
            q[i++] = Character.toLowerCase(ch);
            int j = 0;
            for (int k = i; k < n; k++, j++) { 
                if (p[j] != q[k]) { 
                    continue search;
                }
            }
            for (int k = 0; j < n; k++, j++) { 
                if (p[j] != q[k]) { 
                    continue search;
                }
            }
            return fromPos - n;
        }
        return -1;
    }

    public int addBookmark(Bookmark bm) {
        byte[] record = bm.pack();
        try { 
            bm.id = store.addRecord(record, 0, record.length);
        } catch (Exception x) {}

        return insertBookmark(bm);
    }

    public void updateBookmark(Bookmark bm) { 
        byte[] record = bm.pack();
        try { 
            store.setRecord(bm.id, record, 0, record.length);
        } catch (Exception x) {}
    }        

    public void removeBookmark(int i) { 
        try { 
            store.deleteRecord(((Bookmark)bookmarks.elementAt(i)).id);
        } catch (Exception x) {}
        bookmarks.removeElementAt(i);
    }        
        


    public void saveCurrentPosition() { 
        byte[] record = new byte[RECORD_SIZE];
        record[FLAGS_OFFS] = (byte)flags;
        record[OVERLAP_OFFS] = (byte)overlap;
        record[FONT_OFFS] = (byte)font;
        Converter.packInt(record, POSITION_OFFS, currPagePos);
        try { 
            store.setRecord(1, record, 0, record.length);
        } catch (Exception x) {}
    }


    private void openStore() {
        try { 
            store = RecordStore.openRecordStore(STORE_NAME, true);
            int last = store.getNextRecordID();
            if (last > 1) { 
                byte[] record = store.getRecord(1);
                flags = record[FLAGS_OFFS];
                font = record[FONT_OFFS];
                overlap = record[OVERLAP_OFFS];
                currPagePos = Converter.unpackInt(record, POSITION_OFFS);
                for (int i = 2; i < last; i++) { 
                   try { 
                       insertBookmark(new Bookmark(i, store.getRecord(i)));
                   } catch (InvalidRecordIDException x) {}
               }
            } else { 
                byte[] record = new byte[RECORD_SIZE];
                flags = SettingsForm.SLIDER_HORIZONTAL|SettingsForm.SHOW_CLOCK;
                overlap = DEFAULT_OVERLAP;
                font = SettingsForm.SMALL_FONT;
                record[FLAGS_OFFS] = (byte)flags;
                record[OVERLAP_OFFS] = (byte)overlap;
                record[FONT_OFFS] = (byte)font;
                store.addRecord(record, 0, record.length);
                addBookmark(new Bookmark(0, Locale.BEGIN));

                try {
                    DataInputStream content = new DataInputStream(this.getClass().getResourceAsStream("/content"));
                    if (content != null) { 
                        while (true) {
                            int pos = content.readInt();
                            int len = content.readByte() & 0xFF;
                            byte[] nameBytes = new byte[len];
                            content.readFully(nameBytes);
                            char[] nameChars = new char[len];
                            for (int i = 0; i < len; i++) { 
                                nameChars[i] = Locale.getChar(nameBytes[i]);
                            }
                            addBookmark(new Bookmark(pos, new String(nameChars)));
                        }
                    }
                } catch (Exception x) {}

                if (fileLength != 0) { 
                    addBookmark(new Bookmark(Integer.MAX_VALUE, Locale.END));
                }
            }
        } catch(Exception x) {
        }
    }

    private void closeStore() {
        if (store != null) { 
            try {                
                store.closeRecordStore();
                store = null;
            } catch (Exception x) {}
        }
    }

    public BookReader() {
        String bookSize = getAppProperty("book.size");
        if (bookSize != null) { 
            fileLength = Integer.parseInt(bookSize);
        } else { 
            try {
                DataInputStream cfg = new DataInputStream(this.getClass().getResourceAsStream("/config"));
                fileLength = cfg.readInt();
            } catch (IOException e) {}        
        }
        calendar = Calendar.getInstance();        
        buf = new byte[BUF_SIZE];
        rewind ();
        fillBuffer();
        bookmarks = new Vector();
        openStore();
    }
        
    String getTime() { 
        calendar.setTime(new Date());
        int hour = calendar.get(Calendar.HOUR);
        String h = "0" + (calendar.get(Calendar.AM_PM) == Calendar.PM && hour < 12
                          ? 12 + hour : hour);
        String m = "0" + calendar.get(Calendar.MINUTE);
        return h.substring(h.length()-2) + ':' + m.substring(m.length()-2);
    }


    private void fillBuffer() {
        int offs = 0;
        try { 
            while (offs < buf.length) { 
                int rc = in.read(buf, offs, buf.length - offs);
                if (rc <= 0) { 
                    break;
                }
                offs += rc;
            }
        } catch (IOException x) {}
        
        if (offs >= 0 && offs < BUF_SIZE) { 
            buf[offs] = '\0';
        }
    }

    private boolean moveTo(int pos) {        
        // seek is needed
        int offs = pos <= BUF_SIZE/2 ? 0 : pos - BUF_SIZE/2;
        if (pos < bufPos) { 
            rewind();
        } else { 
            bufPos += BUF_SIZE;
            offs -= bufPos;
        }
        if (offs > 0) { 
            long skipped = 0;
            try { 
                skipped = in.skip(offs);
            } catch (IOException x) {}

            if (skipped < offs) { 
                //System.out.println("EOF!!!!!");
                return false;
            }
            bufPos += offs;
        }
        fillBuffer();
        return true;
    }

    private void rewind() { 
        bufPos = 0;
        if (in != null) { 
            try { 
                in.close();
            } catch (IOException x) {}
        }
        in = getClass().getResourceAsStream(BOOK);
    }

    public char getChar(int pos) { 
        if (pos < bufPos) { 
            if (!moveTo(pos)) { 
                return '\0';
            }               
        } else if (pos >= bufPos + BUF_SIZE) {
            if (pos == bufPos + BUF_SIZE) {
                if (!moveForward()) { 
                    return '\0';
                }
            } else { 
                if (!moveTo(pos)) { 
                    return '\0';
                }
            }
        }
        return Locale.getChar(buf[pos - bufPos]);
    }
    
    private boolean moveForward() {      
        System.arraycopy(buf, BUF_SIZE/2, buf, 0, BUF_SIZE/2);
        bufPos += BUF_SIZE/2;
        int offs = 0;
        try { 
            while (offs < BUF_SIZE/2) { 
                int rc = in.read(buf, BUF_SIZE/2 + offs, BUF_SIZE/2 - offs);
                if (rc <= 0) { 
                    if (offs == 0) { 
                        return false;
                    }
                    break;
                }
                offs += rc;
            }
        } catch (IOException x) {}

        if (offs < BUF_SIZE/2) {
            buf[BUF_SIZE/2 + offs] = 0;
        }
        return true;
    }

    public void error(String message, Displayable parent) { 
        Alert a = new Alert(Locale.ERROR, message, null, AlertType.ERROR);
        a.setTimeout(Alert.FOREVER);
        Display.getDisplay(this).setCurrent(a, parent);
    }


    protected void destroyApp(boolean unconditional) { 
        saveCurrentPosition();
        closeStore();
    }    

    protected  void pauseApp() {
        saveCurrentPosition();
    }

    protected void startApp() 
    {
        Display.getDisplay(this).setCurrent(new ReaderCanvas(this));
    }

    public void quit() { 
        destroyApp(true);
        notifyDestroyed();
    }

    InputStream in;

    static final int POSITION_OFFS = 0;
    static final int FLAGS_OFFS    = 4;
    static final int FONT_OFFS     = 5;
    static final int OVERLAP_OFFS  = 6;
    static final int RECORD_SIZE   = 8;

    Vector      bookmarks;
    RecordStore store;
    int         bufPos;
    int         currPagePos;
    int         fileLength;
    int		overlap;
    int         font;
    int         flags;
    byte []     buf;
    Calendar    calendar;
}

