import javax.microedition.lcdui.*;
import com.nokia.mid.ui.FullCanvas;
import java.util.*;

public class ReaderCanvas extends FullCanvas {
    BookReader reader;
    int        screenWidth;
    int        screenHeight;
    int        fontHeight;
    int        screenLines;
    int        nLines;
    int[]      linePos;
    char[]     linePrevChar;
    char       prevPageChar;
    IFont      font;
    Font       clockFont;
    Image      imgBuf;
    int        spaceWidth;
    int        hyphenWidth;
    String     searchPattern;
    int        searchPos;
    String     defaultBookmark;
    boolean    initialized;
    Timer      repeatTimer;
    RepeatTask repeatTask;

    static final int WHITE = 0xffffff;
    static final int BLACK = 0;
    static final int SLIDER_COLOR = 0x8080ff;
    static final int SLIDER_POS_COLOR = 0xff0000;
    static final int CLOCK_COLOR = 0x88ffff;

    static final int MIN_SLIDER_LENGTH = 3;
    static final int MAX_BOOKMARK_LENGTH = 16;

    static final int REPEAT_DELAY = 1000;
    static final int REPEAT_PERIOD = 100;

    static final int INDENTATION = 3;
    
    class RepeatTask extends TimerTask { 
        RepeatTask(int keyCode) { 
            this.keyCode = keyCode;
        }

        public void run() { 
            keyAction(keyCode);
        }

        int keyCode;
    }

    ReaderCanvas(BookReader reader) {
        this.reader = reader;
        prevPageChar = '\n';
        searchPos = -1;
        initialized = false;
        repeatTimer = new Timer();
        clockFont = Font.getFont(Font.FACE_MONOSPACE, Font.STYLE_BOLD, Font.SIZE_LARGE);
    }

    void resetFont() 
    { 
        switch (reader.font) { 
        case SettingsForm.SMALL_FONT:
            font = new SystemFontWrapper(Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_SMALL));
            break;
        case SettingsForm.MEDIUM_FONT:
            font = new SystemFontWrapper(Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_MEDIUM));
            break;
        case SettingsForm.LARGE_FONT:
            font = new SystemFontWrapper(Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_LARGE));
            break;
        default:
            font = new ImageFont("/fonts/" + SettingsForm.fontName[reader.font]);
        }
        fontHeight = font.getHeight() - reader.overlap;
        spaceWidth = font.charWidth(' ');
        hyphenWidth = font.charWidth(Locale.HYPHEN);

        screenHeight = getHeight();
        if ((reader.flags & SettingsForm.SLIDER_HORIZONTAL) != 0) { 
            screenHeight -= 1;
        }
        screenWidth = getWidth();
        if ((reader.flags & SettingsForm.SLIDER_VERTICAL) != 0) { 
            screenWidth -= 1;
        }
        screenLines = screenHeight/fontHeight;
        linePos = new int[screenLines+1];
        linePrevChar = new char[screenLines+1];
        initialized = true;
    }

    static void insertSpaces(StringBuffer buf, int nSpaces) { 
        if (nSpaces > 0) { 
            int i = 0;
            int n = buf.length();
            while (i < n && buf.charAt(i) == ' ') { 
                i += 1;
            }
            int nSeparators = 0;
            while (++i < n) { 
                if (buf.charAt(i) ==  ' ') { 
                    nSeparators += 1;
                }
            }
            if (nSeparators != 0) { 
                int ns1 = nSpaces / nSeparators;
                int ns2 = nSpaces % nSeparators;
                do { 
                    if (buf.charAt(--i) ==  ' ') { 
                        for (int j = 0; j < ns1; j++) { 
                            buf.insert(i, ' ');
                        }
                        if (ns2 != 0) { 
                            buf.insert(i, ' ');
                            ns2 -= 1;
                        }
                        nSeparators -= 1;
                    }
                } while (nSeparators != 0);
            }
        }
    }

    private static boolean isSeparator(char ch) { 
        return ch == ' ' || ch == '!' || ch == '.' || ch == ',' || ch == ';' || ch == '?' 
            || ch == '-' || ch == '+' || ch == '=' || ch == ')' || ch == '}' || ch == ']' 
            || ch == '\\' || ch == '/' || ch == ':' || ch == '>';
    }

    int splitWord(StringBuffer buf, int n, int width) { 
        char c1 = buf.charAt(n-1);
        char c2 = buf.charAt(n-2);
        if (isSeparator(c2)) { 
            if (c2 == ' ') { 
                buf.delete(n-2, buf.length());
                width -= spaceWidth + font.charWidth(c1);
            } else { 
                buf.delete(n-1, buf.length());
                width -= font.charWidth(c1);
            }
            insertSpaces(buf, (screenWidth - width)/spaceWidth);
            return 1;
        }
        int i = Locale.split(buf, n);
        if (i < 0) {         
            for (i = n-2; i > 0 && !isSeparator(buf.charAt(i-1)); i--);
        }
        if (i <= 0) { 
            i = n-2;
        } 
        int rest = n-i;
        boolean insertHyphen = true;
        if (i >= 1 && isSeparator(buf.charAt(i-1))) { 
            i -= 1;
            insertHyphen = false;
        }             
        for (int j = i; j < n; j++) { 
            width -= font.charWidth(buf.charAt(j));
        }
        buf.delete(i, buf.length());
        if (insertHyphen) { 
            buf.insert(i, Locale.HYPHEN);
            width += hyphenWidth;
        }
        insertSpaces(buf, (screenWidth - width)/spaceWidth);
        return rest;
    }

    protected synchronized void paint(Graphics g) {
        if (!initialized) { 
            resetFont();
        }

        g.setColor((reader.flags & SettingsForm.INVERSE) != 0 ? BLACK : WHITE); 
        g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
        if (reader.font <= SettingsForm.LARGE_FONT && (reader.flags &  SettingsForm.SHOW_CLOCK) != 0) 
        { 
            g.setColor(CLOCK_COLOR); 
            g.setFont(clockFont); 
            g.drawString(reader.getTime(), screenWidth, 0, Graphics.TOP|Graphics.RIGHT);
        }
        g.setColor((reader.flags & SettingsForm.INVERSE) != 0 ? WHITE : BLACK); 
        int pos = reader.currPagePos;
        char prevChar = prevPageChar;
        boolean eof = false;
        nLines = 0;     
        defaultBookmark = null;
        for (int i = 0; !eof && i < screenLines; i++) { 
            linePos[i] = pos;
            linePrevChar[i] = prevChar;
            nLines += 1;
            int width = 0;
            StringBuffer buf = new StringBuffer();
            do { 
                char ch = reader.getChar(pos++);
                if (ch == ' ') { 
                    if (prevChar == '\n') { 
                        if (buf.length() < INDENTATION) { 
                            buf.append(ch); 
                            width += spaceWidth;
                        }
                        continue;
                    } else if (prevChar == ' ' || buf.length() == 0) { 
                        continue;
                    }
                } else if (ch == '\0') {
                    eof = true;
                    break;
                } else if (ch == '\n') {
                    ch = reader.getChar(pos);
                    if (ch == ' ' || ch == '\n') { 
                        prevChar = '\n';
                        break;
                    }
                    ch =  ' ';
                }
                buf.append(ch);
                width += font.charWidth(ch);
                prevChar = ch;
            } while (width <= screenWidth);
            
            if (defaultBookmark == null && linePrevChar[i] == '\n') { 
                String bm = buf.toString().trim();
                if (bm.length() > 0) { 
                    if (bm.length() > MAX_BOOKMARK_LENGTH) { 
                        int j = MAX_BOOKMARK_LENGTH;
                        while (j > 0 && bm.charAt(j) != ' ') { 
                            j -= 1;
                        }
                        bm = bm.substring(0, j);
                    }
                    defaultBookmark = bm;
                }
            }
            if (width > screenWidth) { 
                if (prevChar == ' ' || prevChar == '\n') { 
                    buf.deleteCharAt(buf.length()-1);
                    width -= spaceWidth;
                    insertSpaces(buf, (screenWidth - width) / spaceWidth);
                } else { 
                    int n = buf.length();
                    if (!eof) { 
                        int p = pos;
                        char ch;
                        while ((ch = reader.getChar(p++)) > ' ') { 
                            buf.append(ch);
                        }
                    }
                    pos -= splitWord(buf, n, width);
                }
            }
            font.drawString(g, 0, i*fontHeight, buf.toString());
        }
        linePos[nLines] = pos;
        linePrevChar[nLines] = prevChar;

        if (reader.fileLength != 0) { 
            if ((reader.flags & SettingsForm.SLIDER_HORIZONTAL) != 0) { 
                int sliderLength = (pos - reader.currPagePos)*screenWidth/reader.fileLength;
                if (sliderLength < MIN_SLIDER_LENGTH) { 
                    sliderLength = MIN_SLIDER_LENGTH;
                }
                int sliderPos = reader.currPagePos*screenWidth/reader.fileLength;
                g.setColor(SLIDER_COLOR);
                g.drawLine (0, screenHeight, screenWidth, screenHeight);
                g.setColor(SLIDER_POS_COLOR);              
                g.drawLine(sliderPos, screenHeight, sliderPos+sliderLength, screenHeight);
            } 
            if ((reader.flags & SettingsForm.SLIDER_VERTICAL) != 0) { 
                int sliderLength = (pos - reader.currPagePos)*screenHeight/reader.fileLength;
                if (sliderLength < MIN_SLIDER_LENGTH) { 
                    sliderLength = MIN_SLIDER_LENGTH;
                }
                int sliderPos = reader.currPagePos*screenHeight/reader.fileLength;
                g.setColor(SLIDER_COLOR);
                g.drawLine(screenWidth, 0, screenWidth, screenHeight);
                g.setColor(SLIDER_POS_COLOR);              
                g.drawLine(screenWidth, sliderPos, screenWidth, sliderPos+sliderLength);
            }
        }
    }

    void moveBackward(int shift) { 
         int pos = reader.currPagePos;
         int width = 0;
         prevPageChar = '\n';
         while (--pos >= 0 && shift > 0) { 
             char ch = reader.getChar(pos);
             if (ch == '\n') { 
                 width = 0;
                 shift -= 1;
                 prevPageChar = '\n';
             } else { 
                 width += font.charWidth(ch);
                 if (width > screenWidth) { 
                     width = 0;
                     shift -= 1;
                     prevPageChar = ' ';
                 }
             }
         }
         reader.currPagePos = pos+1;
    }    

    void moveForward(int shift) { 
        if (nLines == screenLines) { 
            if (shift > screenLines) {
                shift = screenLines;
            }
            reader.currPagePos = linePos[shift];
            prevPageChar = linePrevChar[shift];                
        }
    }

    public void setPosition(int pos) { 
        if (pos == Integer.MAX_VALUE) { 
            reader.currPagePos = reader.fileLength;
            moveBackward(screenLines/2);
        } else { 
            reader.currPagePos = pos;
        }
    }
        
    public void keyPressed(int keyCode) {
        if (repeatTask != null) { 
            repeatTask.cancel();
        }
        repeatTask = new RepeatTask(keyCode);
        repeatTimer.schedule(repeatTask, REPEAT_DELAY, REPEAT_PERIOD);
    }

    public void keyReleased(int keyCode) {
        keyAction(keyCode);
        if (repeatTask != null) { 
            repeatTask.cancel();
            repeatTask = null;
        }
    }

    public synchronized void keyAction(int keyCode)
    {
        switch (keyCode) 
        {
        case KEY_NUM1:
        case KEY_NUM2:
        case KEY_NUM3:
        case KEY_NUM4:
        case KEY_NUM5:
        case KEY_NUM6:
        case KEY_NUM7:
        case KEY_NUM8:
        case KEY_NUM9:
            moveForward(keyCode - KEY_NUM0);
            break;            
        case KEY_NUM0:
            new BookmarkList(this, this);
            return;
        case KEY_STAR:
            if (searchPos >= 0) { 
                searchPos = reader.search(searchPos+searchPattern.length(), searchPattern);
                if (searchPos >= 0) { 
                    reader.currPagePos = searchPos;
                    break;
                } else { 
                    reader.error(Locale.NOT_FOUND, this);
                    return;
                }
            }
            // no break
        case KEY_POUND:
            new Menu(this);
            return;
        default:
            switch (getGameAction(keyCode)) { 
            case LEFT:
                moveBackward(1);
                break;
            case RIGHT:
                moveForward(1);
                break;
            case UP:
                moveBackward(screenLines);
                break;
            case DOWN:
                moveForward(screenLines);
                break;
            default:
                return;
            }
        }
        repaint();
        reader.saveCurrentPosition();
    }
}
